from pathlib import Path
from typing import List, Optional, Tuple

from pydantic import BaseModel, Field, ValidationError
from tinydb import Query, TinyDB
from xdg_base_dirs import xdg_data_home


class DuplicateTemplateId(Exception):
    """When a template with that id already exists in the database"""

    pass


class TemplateNotFound(Exception):
    """When we can't find a template in the db"""

    pass


class CorruptedTemplate(Exception):
    """When we're not able to construct a Template from a DB record"""


class Template(BaseModel):
    """
    The model for the templates copied over to the clipboard. Contains the
    template identifier as well as the template string.

    Note that the id may contain only alphanumeric characters or `-` and `_`
    """

    identifier: str = Field(min_length=1, pattern=r"^[a-zA-Z0-9_-]+$")
    template: str = Field(min_length=1)

    def display(self) -> Tuple[str, str]:
        """
        Display returns the identifier and template marked up for insertion into
        a rich.print() statement
        """
        template_formatted = self.template.replace(
            r"{}", r"[magenta]{}[/magenta]"
        )
        return (
            f"[bold blue]{self.identifier}[/bold blue]",
            f"{template_formatted}",
        )


class DB:
    """
    Convenience interface from the json "database" to Templates and back
    """

    data_dir: Path = xdg_data_home() / "tmplcl"
    db: TinyDB

    def __init__(self, data_dir: Optional[Path] = None):
        if data_dir:  # testing helper
            self.data_dir = data_dir

        if not self.data_dir.is_dir():
            self.data_dir.mkdir()

        self.db = TinyDB(self.data_dir / "data.json")

    def insert(self, template: Template):
        if template not in self.get_all():
            self.db.insert(template.model_dump())
        else:
            raise DuplicateTemplateId

    def get(self, identifier) -> Tuple[Template, int]:
        match = self.db.search(Query().identifier == identifier)
        if match:
            try:
                return Template(**match[0]), match[0].doc_id
            except ValidationError:
                raise CorruptedTemplate
        else:
            raise TemplateNotFound

    def get_all(self) -> List[Template]:
        return [Template(**item) for item in self.db.all()]

    def update(self, identifier: str, template_str: str):
        template, id = self.get(identifier)
        updated = Template(
            identifier=template.identifier, template=template_str
        )
        self.db.update(updated.model_dump(), doc_ids=[id])

    def delete(self, identifier):
        self.db.remove(Query().identifier == identifier)



from typing import Optional
from typing_extensions import Annotated
import typer
from pydantic import ValidationError

from tmplcl.commands import (
    add_template,
    copy_template,
    delete_template,
    list_templates,
    show_template,
    update_template,
)
from tmplcl.models import DB, CorruptedTemplate, TemplateNotFound

app = typer.Typer()
db = DB()


@app.command()
def copy(template: str):
    """Copies the requested template to your clipboard"""
    try:
        copy_template(template, db)
    except TemplateNotFound:
        print(
            "[red bold]Error[/red bold]: Unable to find the requested "
            f'template "[magenta bold]{template}[/magenta bold]" in '
            "the database."
        )
        raise typer.Exit(code=1)

    except CorruptedTemplate:
        print(
            "[red bold]Error[/red bold]: The requested template "
            f'"[bold]{template}[/bold]" appears to have been corrupted.'
            f"Please either remove or fix the entry in [italic]{db.data_dir}"
            "/data.json[/italic]."
        )
        raise typer.Exit(code=1)


@app.command()
def delete(template: str):
    """Deletes the template with the provided identifier"""
    delete_template(template, db)


@app.command()
def add(
    template_name: Annotated[
        str,
        typer.Argument(
            help="The name that you'll use to retreive this template"
        ),
    ],
    template_string: Annotated[
        str,
        typer.Argument(
            help="The actual string that will be copied to your clipboard"
        ),
    ],
):
    """Adds a template with the provided identifier and string"""
    add_template(template_name, template_string, db)


@app.command()
def list(
    show_chars: Annotated[
        Optional[int],
        typer.Option(
            help=(
                "An interger value for the number of characters "
                "you want to see in each preview"
            )
        ),
    ] = None,
):
    """
    Lists all available templates, including a preview of each
    """
    if show_chars:
        list_templates(db, show_chars)
    else:
        list_templates(db)


@app.command()
def show(template: str):
    """
    Displays the full text of a given template
    """
    try:
        show_template(template, db)
    except TemplateNotFound:
        print(
            "[red bold]Error[/red bold]: Unable to find the requested "
            f'template "[magenta bold]{template}[/magenta bold]" in '
            "the database."
        )
        raise typer.Exit(code=1)
    except CorruptedTemplate:
        print(
            "[red bold]Error[/red bold]: The requested template "
            f'"[bold]{template}[/bold]" appears to have been corrupted.'
            f"Please either remove or fix the entry in [italic]{db.data_dir}"
            "/data.json[/italic]."
        )
        raise typer.Exit(code=1)


@app.command()
def update(
    template: Annotated[
        str,
        typer.Argument(help="The name of the template you want to update"),
    ],
    template_string: Annotated[
        str,
        typer.Argument(
            help="The string that you want to update this template to"
        ),
    ],
):
    """Updates a given template with a new string"""
    try:
        update_template(template, template_string, db)
    except TemplateNotFound:
        print(
            "[red bold]Error[/red bold]: Unable to find the requested "
            f'template "[magenta bold]{template}[/magenta bold]" in '
            "the database."
        )
        raise typer.Exit(code=1)
    except ValidationError:
        print("[red bold]Error[/red bold]: Invalid replacement template string")
        raise typer.Exit(code=1)



from typing import Optional

import pyperclip
from rich import print
from typing_extensions import Annotated

from tmplcl.models import DB, Template
from tmplcl.utils import SKIP_OPTION


# CREATE
def add_template(identifier: str, template_str: str, db: DB) -> None:
    """
    Adds a new template to the database
    """
    db.insert(Template(identifier=identifier, template=template_str))


# READ
def copy_template(
    template: str, db: Annotated[Optional[DB], SKIP_OPTION] = None
) -> None:
    """
    Finds a template by its id and copies the resultant string to the clipboard
    """
    if not db:
        db = DB()

    retrieved, _ = db.get(template)
    pyperclip.copy(retrieved.template)


def list_templates(db: DB, truncate: int = 50) -> None:
    """
    Lists all available templates, with a truncated preview of template string
    """
    for template in db.get_all():
        id, templ = template.display()
        if len(templ) > truncate:
            print(f"{id}: [italic]{templ[:truncate]}[/italic]...")
        else:
            print(f"{id}: [italic]{templ}[/italic]")


def show_template(identifer: str, db: DB) -> None:
    """
    Shows the full "template" for a given identifier, formatted to highlight any
    internal template string options
    """
    template, _ = db.get(identifer)
    print(f"{template.display()[1]}")


# UPDATE
def update_template(identifier: str, template_str: str, db: DB) -> None:
    """
    Updates a given template id with a new template string
    """
    db.update(identifier, template_str)


# DESTROY
def delete_template(identifier: str, db: DB) -> None:
    """
    Finds a template by its id and deletes it from the database
    """
    db.delete(identifier)



import typer

from tmplcl.app import app
from tmplcl.commands import copy_template


def main():  # pragma: nocover
    """
    Passthrough to run the app
    """
    app()


def copy():  # pragma: nocover
    """
    Passthrough to run just the copy command for convenience
    """
    typer.run(copy_template)
